﻿// Copyright(c) 2015 DatumTron Corp, All Rights Reserved
// Author: Ashraf Azmi

// ReSharper disable SuggestVarOrType_SimpleTypes
// ReSharper disable InconsistentNaming
// ReSharper disable SuggestVarOrType_BuiltInTypes
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using DatumTrom.Datum.AbstractExt;
using DatumTron.Datum.Abstract;
using DatumTron.Datum.Atum;

namespace DatumTron.Apps
{
    public class Program
    {
        static void Main(string[] args)
        {
            Toutorial_1();
            Toutorial_2();

            Console.WriteLine("\n\nPress any key to exit . . .");
            Console.ReadKey(); 
        }

        private static void Toutorial_1()
        {
            #region --- 1. setup --------------------------

            katum thing = datum.setup(new atum());

            #endregion

            #region --- 2.1 adding data: new katums -------

            katum color = thing.get("color"); // or use the ‘|’ as in the following
            katum taste = thing | "taste";
            katum food = thing | "food";
            katum fruit = food | "fruit";
            katum apple = fruit | "apple";
            katum mango = fruit | "mango";
            katum red = color | "red";
            katum sweet = taste | "sweet";

            #endregion

            #region --- 2.2 adding data: attributes -------

            // apple is red
            apple.@is(red); //OR

            // apple can have one color which is red
            apple.set(color, red); //OR
            color[apple] = red; //OR
            color[apple] = color | "red";
            apple += color | red;
            OK(apple.Is(red));

            apple.isnot(red); // apple is not red anymore
            OK(!apple.Is(red));
            apple.set(red); // back to apple is red

            #endregion

            #region --- 3. query --------------------------

            katum red1 = color.find("red");
            OK(red1 == red);
            katum red2 = color & "red";
            OK(red2 == red);
            katum blue = color & "blue";
            OK(blue == null);

            OK(red.Is(color)); // Is
            OK(red.Isa(color)); // Isa

            OK(color.of(apple) == red);
            OK(color[apple] == red);

            #endregion

            #region --- 4. inheritance --------------------          

            katum fujiApple = apple | "fujiApple";
            OK(color[fujiApple] == red); // inherited from apple

            katum darkRed = red | "darkRed";
            apple.set(darkRed);

            OK(color[fujiApple] == darkRed);
            OK(color[fujiApple].Is(red));
            OK(red[fujiApple] == darkRed);

            #endregion

            #region --- 5 Time ----------------------------

            katum pricePerPound = thing | "pricePerPound";
            fujiApple.set(pricePerPound | 1.5);

            fujiApple.now(pricePerPound | 1.3); // price changing over time
            fujiApple.now(pricePerPound | 1);

            OK(pricePerPound[fujiApple] == 1); // price of the current time instance of fujiApple
            OK(pricePerPound[fujiApple.at(0)] == 1.5); //OR just use [0]
            OK(pricePerPound[fujiApple[0]] == 1.5);
            OK(pricePerPound[fujiApple[1]] == 1.3);
            OK(fujiApple.countI == 2); // number of instances under fujiApple

            #endregion

            #region --- 6 Code ----------------------------

            katum banana = fruit | "banana";
            katum weight = thing | "weight";
            apple.set(weight | 1);
            mango.set(color | "green", weight | 0.9);
            banana.set(color | "yellow", weight | 1.1);

            fruit.set(weight | AvgFunc); // <---
            OK(weight[fruit] == 1); // average of 1, .9, 1.1 

            #endregion
        }

        private static void Toutorial_2()
        { 
            #region --- 7 Query  -------------------------

            #region --- 7.1 Setup for Northwind ------------- 

            katum northwindDB = katum.load(new atum(), @"..\..\northwind.datum");
            OK(northwindDB != null);

            datum.thing = (katum<string>)northwindDB;

            katum Table = northwindDB & "Table";    OK(Table != null);
            katum Column = northwindDB & "Column";  OK(Column != null);

            katum Employee = Table & "Employee";    OK(Employee != null);
            katum Order = Table & "Order";          OK(Order != null);
            katum Customer = Table & "Customer";    OK(Customer != null);

            katum EmployeeCols = Column & "Employee";    OK(EmployeeCols != null);
            katum OrderCols = Column & "Order";          OK(OrderCols != null);
            katum CustomerCols = Column & "Customer";    OK(CustomerCols != null);
            katum ShipperCols = Column & "Shipper";      OK(ShipperCols != null);

            #endregion

            #region --- 7.2 Iterate over all employees

            OK(Employee.I != null);
            Console.WriteLine("All Employees");
            foreach (katum employee in Employee.I) 
                Console.WriteLine(employee); // work with each employee

            Console.WriteLine("All Employees again");
            
            // one line version
            Employee.doI(Console.WriteLine);

            #endregion

            #region --- 7.3 Iterate over USA employees

            katum USAEmployees = EmployeeCols & "Country" & "USA";  // O(1)
            OK(USAEmployees!=null);

            Console.WriteLine("USA Employees");
            foreach (katum employee in USAEmployees.I)
                Console.WriteLine(employee);

            // one line version
            (EmployeeCols & "Country" & "USA").doI(Console.WriteLine);

            #endregion

            #region --- 7.4 Calculating TotalPrice using an IFunction class

            katum OrderDetailTable      = Table & "Order Detail";           OK(OrderDetailTable != null);
            katum OrderDetailCols       = Column & "Order Detail";          OK(OrderDetailCols!= null);

            katum OrderDetailPriceCol   = OrderDetailCols & "UnitPrice";    OK(OrderDetailPriceCol!=null);
            katum OrderDetailQtyCol     = OrderDetailCols & "Quantity";     OK(OrderDetailQtyCol != null);
            katum OrderDetailDiscountCol= OrderDetailCols & "Discount";     OK(OrderDetailDiscountCol != null);

            katum OrderDetailTotalPrice = OrderDetailCols | "totalPrice";   OK(OrderDetailTotalPrice != null);
            katum calcTotalPrice        = OrderDetailTotalPrice | new TotalPrice(); OK(calcTotalPrice != null);

            foreach (var orderDetail in OrderDetailTable.I)
            {
                double price    = OrderDetailPriceCol   [orderDetail].asDouble;
                double qty      = OrderDetailQtyCol     [orderDetail].asDouble;
                double discount = OrderDetailDiscountCol[orderDetail].asDouble;
                var totalPrice  = OrderDetailTotalPrice [orderDetail].asDouble;

                Console.WriteLine( "{0} ( price {1} * qty {2} ) - discount {3} \t= totalPrice {4}",
                                    orderDetail, price, qty, discount, totalPrice);
            }

            #endregion

            #region --- 7.5 List employees older than 50 years old

            var EmployeeBD = EmployeeCols & "BirthDate";
            foreach (var employee in Employee.I.Where(e => EmployeeBD[e] < DateTime.Today.AddYears(-50)))
                Console.WriteLine(employee);

            // Shorter alternative
            Employee.doI(Console.WriteLine, e => EmployeeBD[e] < DateTime.Today.AddYears(-50));

            #endregion

            #region --- 7.6 List the order number and employee's last name ...
            // List the order number and employee's last name that 
            // have placed orders to be delivered in Belgium.

            var EmployeeLastNameCol = EmployeeCols & "LastName";
            var BelgiumOrders = OrderCols & "ShipCountry" & "Belgium"; // O(1)
            foreach (var order in BelgiumOrders.I)
                Console.WriteLine("{0} by {1}", order, EmployeeLastNameCol[Employee[order]]);

            #endregion

            #region --- 7.7 Give the order id, employee id ...
            // Give the order id, employee id and the customer id for 
            // orders that are sent by the company ‘Speedy Express’ to customers 
            // who live in Buenos Aires.

            var SpeedyExpressShipper = ShipperCols & "CompanyName" & "Speedy Express"; // O(1)
            var BuenosAiresCustomers = CustomerCols & "City" & "Buenos Aires"; // O(1)

            var O1 = Order & SpeedyExpressShipper.I; // orders sent by Speedy Express            
            var O2 = Order & BuenosAiresCustomers.I; // orders bought by Brussels customers

            foreach (var order in O1.Intersect(O2))  
                Console.WriteLine("{0} {1} {2}", order, Employee[order], Customer[order]);
            #endregion

            #region --- 7.8 List products that were bought or sold by people who live in London. 

            var Category    = Table & "Category";
            var Product     = Table & "Product";         OK(Product!=null);
            var OrderDetail = Table & "Order Detail";    OK(OrderDetail!=null);

            var customerCityLondon = Column & "Customer" & "City" & "London";
            var employeeCityLondon = Column & "Employee" & "City" & "London";

            var OD1 = OrderDetail & (Order & (Customer & customerCityLondon));
            var OD2 = OrderDetail & (Order & (Employee & employeeCityLondon));

            foreach (var orderDetail in OD1.Union(OD2))
                Console.WriteLine( Product[orderDetail] );

            #endregion

            #endregion Query

            #region --- 8 Finding Patterns ---------------

            Console.WriteLine("\n--- Patterns: Any pattern in customer City and orderd product Category?\n");

            const int minSupport = 10;    // minimum 10 orders 
            const int minConfidence = 20; // minimum 20% probability

            // Customer City and Product Category 
            var categoryName = Column & "Category" & "CategoryName";

            foreach (var city in (Column & "Customer" & "City").I)
            {
                var categories = Category[Product[ OrderDetail & (Order & (Customer & city)), distinct:false ], distinct: false];
                #region debug code
                var categorie2 = VPatterns(city, new[] { Customer, Order, OrderDetail }, new[] { Product, Category });
                var dbgCat1 = categories.ToArray();
                var dbgCat2 = categorie2.ToArray();
                Debug.Assert(dbgCat1.Length == dbgCat2.Length);
                #endregion
                int total;
                var freqDict = FreqDict(categories, out total);
                PrintPattern(city, categoryName, freqDict, total, minSupport, minConfidence);
                #region debug code
                /*
                var dbg0 = (Customer & city).ToArray();
                var dbg1 = (Order & (Customer & city)).ToArray();
                var dbg2 = (OrderDetail & (Order & (Customer & city))).ToArray();
                var dbg3 = Product[OrderDetail & (Order & (Customer & city)), distinct: false].ToArray();
                Debug.Assert(dbg2.Count() == dbg3.Count());              

                var dbg4 = categories.ToArray();
                Debug.Assert(dbg2.Count() == dbg4.Count());
                Debug.Assert(dbg2.Count() == total);
                */
                #endregion
            }

            Console.WriteLine("\n--- Patterns: Any patterns in Employee selling product Category?\n");

            // Employee and Product Category
            foreach (var employee in (Table & "Employee").I0)
            {
                var categories = Category[Product[ OrderDetail & (Order & employee),distinct: false], distinct: false];
                int total;
                var freqDict = FreqDict(categories, out total);
                #region to help understand, put a break point and view dbg variables
                
                var dbg1 = (Order & employee).ToArray();
                var dbg2 = (OrderDetail & (Order & employee)).ToArray();
                var dbg3 = Product[OrderDetail & (Order & employee), distinct: false].ToArray();
                Debug.Assert(dbg2.Count() == dbg3.Count());

                var dbg4 = categories.ToArray();
                Debug.Assert(dbg2.Count() == dbg4.Count());
                Debug.Assert(dbg2.Count() == total);
                
                #endregion
                PrintPattern(employee, categoryName, freqDict, total, minSupport, minConfidence);
            }

            // Employee and Product Category
            foreach (var employee in (Table & "Employee").I0)
            {
                var categories = VPatterns(employee, new[] { Order, OrderDetail }, new[] { Product, Category });
                int total;
                var freqDict = FreqDict(categories, out total);
                PrintPattern(employee, categoryName, freqDict, total, minSupport, minConfidence);
            }
            Console.WriteLine("\nEnd Patterns\n");
            
            var patterns1 = VPatterns(Column & "Customer" & "City",
                new[] {Customer, Order, OrderDetail},
                new[] {Product, Category},
                10,
                20);

            foreach( var pattern in patterns1)
                Console.WriteLine("{0} --> {1} support = {2} probability = {3}%", 
                    pattern.Item1, pattern.Item2, pattern.Item3, pattern.Item4);

            var patterns2 = VPatterns(Table & "Employee",
                new[] {Order, OrderDetail},
                new[] {Product, Category},
                10,
                20);

            foreach (var pattern in patterns2)
                Console.WriteLine("{0} --> {1} support = {2} probability = {3}%",
                    pattern.Item1, pattern.Item2, pattern.Item3, pattern.Item4);

            #region debug

            var dbgArr1 = VPatterns(Column & "Customer" & "City",
                new[] {Customer, Order, OrderDetail},
                new[] {Product, Category},
                10,
                20).ToArray();

            var dbgArr2 = VPatterns(Table & "Employee",
                new[] {Order, OrderDetail},
                new[] {Product, Category},
                10,
                20).ToArray();

            #endregion

            #endregion
        }

        #region --- data mining functions

        private static IEnumerable<katum> VPatterns(
            katum condition, 
            katum[] ifPath, 
            katum[] thenPath)
        {
            var eset = ifPath[0] & condition;
            for (var i = 1; i < ifPath.Length; i++)
                eset = ifPath[i] & eset;

            var elist = thenPath[0][eset, distinct: false];
            for (var i = 1; i < thenPath.Length; i++)
                elist = thenPath[i][elist, distinct: false];

            return elist;
        }

        private static IEnumerable<Tuple<katum, katum, int, int>> VPatterns(
            katum conditions, 
            katum[] ifPath, 
            katum[] thenPath, 
            int minSupport, 
            int minProbability)
        {
            foreach (var condition in conditions.I0)
            {
                var patterns = VPatterns(condition, ifPath, thenPath);
                int total;
                var freqDict = FreqDict(patterns, out total);
                foreach (var item in freqDict.Keys)
                {
                    var support = freqDict[item];
                    var probability = (support * 100) / total;

                    if (support >= minSupport && probability > minProbability)
                        yield return new 
                            Tuple<katum, katum, int, int>(condition, item, support, probability);
                }
            }
        }

        #endregion data mining functions

        #region --- Support functions

        static void OK(bool condition) { Debug.Assert(condition); }

        static object AvgFunc(katum container, katum numberAttribute, datum k)
        {
            if (k == null || numberAttribute == null) return null;
            return k.I.Average(i => numberAttribute[i].asDouble);
        }

        /// <summary>
        /// Creates a dictionary where the keys are unique items in the input items, and the key's value is how may times 
        /// the key is found in the input.
        /// </summary>
        /// <typeparam name="T">Type of items</typeparam>
        /// <param name="items">the input items</param>
        /// <param name="total">the count of items</param>
        /// <returns>Frequent Items Dictionary</returns>
        private static Dictionary<T, int> FreqDict<T>(IEnumerable<T> items, out int total)
        {
            var dict = new Dictionary<T, int>(); //How many times each item is found in items
            total = 0;

            foreach (var item in items)
            {
                if (!dict.ContainsKey(item)) dict.Add(item, 1);
                else dict[item]++;
                total++;
            }

            return dict;
        }

        private static void PrintPattern(katum condition, katum conclusion, Dictionary<katum, int> freqDict, int total, int minSupport, int minConfidence)
        {
            foreach (var item in freqDict.Keys)
            {
                var support = freqDict[item];
                var confidence = (support*100)/total;

                if (support >= minSupport && confidence > minConfidence)
                    Console.WriteLine("{0} -> {1}% {2} -- support = {3}",
                        condition, confidence, conclusion[item], support);
            }
        }

        #endregion
    }

    class TotalPrice :IFunction
    {
        //public function Function => TotalPriceFunc; // c# 6.0
        public function Function { get { return TotalPriceFunc; } }

        static katum _orderDetailPriceCol;
        static katum _orderDetailQtyCol;
        static katum _orderDetailDiscountCol;
        
        static private void Initialize()
        {          
            katum Cols = datum.thing & "Column"; OK(Cols != null);
            
            katum OrderDetailCols = datum.thing & "Column" & "Order Detail"; OK(OrderDetailCols != null);

            _orderDetailPriceCol = OrderDetailCols & "UnitPrice"; OK(_orderDetailPriceCol != null);
            _orderDetailQtyCol = OrderDetailCols & "Quantity"; OK(_orderDetailQtyCol != null);
            _orderDetailDiscountCol = OrderDetailCols & "Discount"; OK(_orderDetailDiscountCol != null);
        }

        static object TotalPriceFunc(katum container, katum totalPrice, datum orderDetail)
        {
            if (orderDetail == null || totalPrice == null) return null;

            if(_orderDetailDiscountCol==null) Initialize();
            if (_orderDetailDiscountCol == null) return null;

            var discount = (_orderDetailDiscountCol[orderDetail]).asDouble;
            return (_orderDetailPriceCol[orderDetail].asDouble * _orderDetailQtyCol[orderDetail].asDouble) * (1-discount);
        }

        static void OK(bool cond) { Debug.Assert(cond); }
    }
}
